/*--------------------------------------------------------------------------*/
/*--- ALBERTA:  an Adaptive multi Level finite element toolbox using     ---*/
/*---           Bisectioning refinement and Error control by Residual    ---*/
/*---           Techniques for scientific Applications                   ---*/
/*---                                                                    ---*/
/*--- file: lagrange_3.c                                                 ---*/
/*---                                                                    ---*/
/*--- description: implementation of the basis functions                 ---*/
/*---              lagrange4 in 1d                                       ---*/
/*---                                                                    ---*/
/*--- created by: kgs on host enigma                                     ---*/
/*---           at 16:50 on 28 of March 2003                             ---*/
/*--------------------------------------------------------------------------*/
/*---                                                                    ---*/
/*--- authors:   Alfred Schmidt                                          ---*/
/*---            Zentrum fuer Technomathematik                           ---*/
/*---            Fachbereich 3 Mathematik/Informatik                     ---*/
/*---            Universitaet Bremen                                     ---*/
/*---            Bibliothekstr. 2                                        ---*/
/*---            D-28359 Bremen, Germany                                 ---*/
/*---                                                                    ---*/
/*---            Kunibert G. Siebert                                     ---*/
/*---            Institut fuer Mathematik                                ---*/
/*---            Universitaet Augsburg                                   ---*/
/*---            Universitaetsstr. 14                                    ---*/
/*---            D-86159 Augsburg, Germany                               ---*/
/*---                                                                    ---*/
/*---    http://www.alberta-fem.de/                                      ---*/
/*---                                                                    ---*/
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#include <alberta.h>

/*--------------------------------------------------------------------------*/
/*---  basisfunction 0 located at vertex 0                               ---*/
/*--------------------------------------------------------------------------*/

static REAL phi3_0(const REAL lambda[DIM+1])
{
  return((4.5*(lambda[0] - 1.0)*lambda[0] + 1.0)*lambda[0]);
}

static const REAL *grd_phi3_0(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0};
  grd[0] = (13.5*lambda[0] - 9.0)*lambda[0] + 1;
  return((const REAL *) grd);
}

static const REAL (*D2_phi3_0(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0.0}};
  D2[0][0] = 27.0*lambda[0] - 9.0;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*---  basisfunction 1 located at vertex 1                               ---*/
/*--------------------------------------------------------------------------*/

static REAL phi3_1(const REAL lambda[DIM+1])
{
  return((4.5*(lambda[1] - 1.0)*lambda[1] + 1.0)*lambda[1]);
}

static const REAL *grd_phi3_1(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0};
  grd[1] = (13.5*lambda[1] - 9.0)*lambda[1] + 1;
  return((const REAL *) grd);
}

static const REAL (*D2_phi3_1(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0.0}};
  D2[1][1] = 27.0*lambda[1] - 9.0;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*---  basisfunction 2, first at center                                  ---*/
/*--------------------------------------------------------------------------*/

static REAL phi3_2(const REAL lambda[DIM+1])
{
  return((13.5*lambda[0] - 4.5)*lambda[0]*lambda[1]);
}

static const REAL *grd_phi3_2(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0};
  grd[0] = (27.0*lambda[0] - 4.5)*lambda[1];
  grd[1] = (13.5*lambda[0] - 4.5)*lambda[0];
  return((const REAL *) grd);
}

static const REAL (*D2_phi3_2(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0.0}};
  D2[0][0] = 27.0*lambda[1];
  D2[0][1] = D2[1][0] = 27.0*lambda[0] - 4.5;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*---  basisfunction 3, second at center                                 ---*/
/*--------------------------------------------------------------------------*/

static REAL phi3_3(const REAL lambda[DIM+1])
{
  return((13.5*lambda[1] - 4.5)*lambda[1]*lambda[0]);
}

static const REAL *grd_phi3_3(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0};
  grd[0] = (13.5*lambda[1] - 4.5)*lambda[1];
  grd[1] = (27.0*lambda[1] - 4.5)*lambda[0];
  return((const REAL *) grd);
}

static const REAL (*D2_phi3_3(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0.0}};
  D2[0][1] = D2[1][0] = 27.0*lambda[1] - 4.5;
  D2[1][1] = 27.0*lambda[0];
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------*/
/*--- function for accessing local DOFs on an element              ---*/
/*--------------------------------------------------------------------*/

static const DOF *get_dof_indices3(const EL *el, const DOF_ADMIN *admin,
                                  DOF *idof)
{
  static DOF  dof_vec[4];
  DOF         *rvec = idof ? idof : dof_vec;
  int         ibas = 0, i, j, n0, node;
  DOF         **dof = el->dof;

/*--------------------------------------------------------------------*/
/*--- DOFs at vertices                                             ---*/
/*--------------------------------------------------------------------*/

  node = admin->mesh->node[VERTEX];
  n0   = admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[ibas++] = dof[node+i][n0];

/*--------------------------------------------------------------------*/
/*--- DOFs at center                                               ---*/
/*--------------------------------------------------------------------*/

  node = admin->mesh->node[CENTER];
  n0   = admin->n0_dof[CENTER];
  for (j = 0; j < 2; j++)
    rvec[ibas++] = dof[node][n0+j];


  return(rvec);
}

/*--------------------------------------------------------------------*/
/*--- function for accessing boundary type of DOFs                 ---*/
/*--------------------------------------------------------------------*/

static const S_CHAR *get_bound3(const EL_INFO *el_info, S_CHAR *vec)
{
  FUNCNAME("get_bound3");
  static S_CHAR  my_vec[4];
  S_CHAR         *rvec = vec ? vec : my_vec;
  int            ibas = 0, i, j;

  TEST_FLAG(FILL_BOUND, el_info);

/*--------------------------------------------------------------------*/
/*--- basis functions at vertices                                  ---*/
/*--------------------------------------------------------------------*/

  for (i = 0; i < N_VERTICES; i++)
    rvec[ibas++] = el_info->bound[i];

/*--------------------------------------------------------------------*/
/*--- basis functions at center                                    ---*/
/*--------------------------------------------------------------------*/

  for (j = 0; j < 2; j++)
    rvec[ibas++] = INTERIOR;

  return(rvec);
}

/*--------------------------------------------------------------------*/
/*--- vector of barycentric coordinates                            ---*/
/*--------------------------------------------------------------------*/

static const REAL coords3[4][DIM+1] = 
{
  {1.000000000000000, 0.000000000000000},
  {0.000000000000000, 1.000000000000000},
  {0.666666666666667, 0.333333333333333},
  {0.333333333333333, 0.666666666666667}
};

/*--------------------------------------------------------------------*/
/*--- function for local interpolaton of scalar functions          ---*/
/*--------------------------------------------------------------------*/

static const REAL *interpol3(const EL_INFO *el_info,
                             int no, const int *b_no,
                             REAL (*f)(const REAL_D),
                             REAL *vec)
{
  FUNCNAME("interpol3");
  static REAL   my_vec[4];
  REAL          *rvec = vec ? vec : my_vec;
  int              i, j;
  const PARAMETRIC *parametric = el_info->mesh->parametric;

  if (b_no)
  {
     TEST_EXIT(no > 0 && no <= 4)("not for %d points\n", no);
  }

  if (parametric  &&  parametric->init_element(el_info, parametric))
  {
    REAL_D     world[4];
    REAL       coord_no[4][DIM+1];
    const REAL (*coord)[DIM+1];

    if (b_no)
    {
      for (i = 0; i < no; i++)
      {
        for (j = 0; j <= DIM; j++)
          coord_no[i][j] = coords3[b_no[i]][j];
      }
      coord = (const REAL (*)[DIM+1]) coord_no;
    }
    else
    {
       no = 4;
       coord = coords3;
    }

    (*parametric->coord_to_world)(el_info, nil, no, (void *) coord, world);

    for (i = 0; i < no; i++)
      rvec[i] = (*f)(world[i]);
  }
  else
  {
    TEST_FLAG(FILL_COORDS, el_info);

    if (b_no)
    {
      for (i = 0; i < no; i++)
      {
        const REAL *world = coord_to_world(el_info, coords3[b_no[i]], nil);
        rvec[i] = (*f)(world);
      }
    }
    else
    {
      for (i = 0; i < 4; i++)
      {
        const REAL *world = coord_to_world(el_info, coords3[i], nil);
        rvec[i] = (*f)(world);
      }
    }
  }

  return(rvec);
}
/*--------------------------------------------------------------------*/
/*--- function for local interpolaton of vector functions          ---*/
/*--------------------------------------------------------------------*/

static const REAL_D *interpol_d3(const EL_INFO *el_info,
                                 int no, const int *b_no,
                                 const REAL *(*f)(const REAL_D, REAL_D),
                                 REAL_D *vec)
{
  FUNCNAME("interpol_d3");
  static REAL_D my_vec[4];
  REAL_D        *rvec = vec ? vec : my_vec;
  int              i, j;
  const PARAMETRIC *parametric = el_info->mesh->parametric;

  if (b_no)
  {
     TEST_EXIT(no > 0 && no <= 4)("not for %d points\n", no);
  }

  if (parametric  &&  parametric->init_element(el_info, parametric))
  {
    REAL_D     world[4];
    REAL       coord_no[4][DIM+1];
    const REAL (*coord)[DIM+1];

    if (b_no)
    {
      for (i = 0; i < no; i++)
      {
        for (j = 0; j <= DIM; j++)
          coord_no[i][j] = coords3[b_no[i]][j];
      }
      coord = (const REAL (*)[DIM+1]) coord_no;
    }
    else
    {
       no = 4;
       coord = coords3;
    }

    (*parametric->coord_to_world)(el_info, nil, no, (void *) coord, world);

    for (i = 0; i < no; i++)
      (*f)(world[i], rvec[i]);
  }
  else
  {
    TEST_FLAG(FILL_COORDS, el_info);

    if (b_no)
    {
      for (i = 0; i < no; i++)
      {
        const REAL *world = coord_to_world(el_info, coords3[b_no[i]], nil);
        (*f)(world, rvec[i]);
      }
    }
    else
    {
      for (i = 0; i < 4; i++)
      {
        const REAL *world = coord_to_world(el_info, coords3[i], nil);
        (*f)(world, rvec[i]);
      }
    }
  }

  return((const REAL_D *) rvec);
}
/*--------------------------------------------------------------------*/
/*--- function for accessing a local DOF_INT_VEC                   ---*/
/*--------------------------------------------------------------------*/

static const int *get_int_vec3(const EL *el, const DOF_INT_VEC *dv,
                               int *vec)
{
  FUNCNAME("get_int_vec3");
  static int     my_vec[4];
  int            *rvec = vec ? vec : my_vec;
  int            *v = dv ? dv->vec : nil;
  int            ibas = 0, i, j, n0, node;
  DOF            **dof = el->dof;

  TEST_EXIT(v)("no DOF_INT_VEC dv, or no dv->vec\n");

/*--------------------------------------------------------------------*/
/*--- DOFs at vertices                                             ---*/
/*--------------------------------------------------------------------*/

  node = dv->fe_space->mesh->node[VERTEX];
  n0   = dv->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[ibas++] = v[dof[node+i][n0]];

/*--------------------------------------------------------------------*/
/*--- DOFs at center                                               ---*/
/*--------------------------------------------------------------------*/

  node = dv->fe_space->mesh->node[CENTER];
  n0   = dv->fe_space->admin->n0_dof[CENTER];
  for (j = 0; j < 2; j++)
    rvec[ibas++] = v[dof[node][n0+j]];

  return(rvec);
}

/*--------------------------------------------------------------------*/
/*--- function for accessing a local DOF_REAL_VEC                  ---*/
/*--------------------------------------------------------------------*/

static const REAL *get_real_vec3(const EL *el, const DOF_REAL_VEC *dv,
                                 REAL *vec)
{
  FUNCNAME("get_real_vec3");
  static REAL    my_vec[4];
  REAL           *rvec = vec ? vec : my_vec;
  REAL           *v = dv ? dv->vec : nil;
  int            ibas = 0, i, j, n0, node;
  DOF            **dof = el->dof;

  TEST_EXIT(v)("no DOF_REAL_VEC dv, or no dv->vec\n");

/*--------------------------------------------------------------------*/
/*--- DOFs at vertices                                             ---*/
/*--------------------------------------------------------------------*/

  node = dv->fe_space->mesh->node[VERTEX];
  n0   = dv->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[ibas++] = v[dof[node+i][n0]];

/*--------------------------------------------------------------------*/
/*--- DOFs at center                                               ---*/
/*--------------------------------------------------------------------*/

  node = dv->fe_space->mesh->node[CENTER];
  n0   = dv->fe_space->admin->n0_dof[CENTER];
  for (j = 0; j < 2; j++)
    rvec[ibas++] = v[dof[node][n0+j]];

  return(rvec);
}

/*--------------------------------------------------------------------*/
/*--- function for accessing a local DOF_REAL_D_VEC                ---*/
/*--------------------------------------------------------------------*/

static const REAL_D *get_real_d_vec3(const EL *el, const DOF_REAL_D_VEC *dv,
                                     REAL_D *vec)
{
  FUNCNAME("get_real_d_vec3");
  static REAL_D  my_vec[4];
  REAL_D         *rvec = vec ? vec : my_vec;
  REAL_D         *v = dv ? dv->vec : nil;
  int            ibas = 0, i, j, n, n0, node;
  DOF            **dof = el->dof;

  TEST_EXIT(v)("no DOF_REAL_D_VEC dv, or no dv->vec\n");

/*--------------------------------------------------------------------*/
/*--- DOFs at vertices                                             ---*/
/*--------------------------------------------------------------------*/

  node = dv->fe_space->mesh->node[VERTEX];
  n0   = dv->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
  {
    for (n = 0; n < DIM_OF_WORLD; n++)
      rvec[ibas][n] = v[dof[node+i][n0]][n];
    ibas++;
  }

/*--------------------------------------------------------------------*/
/*--- DOFs at center                                               ---*/
/*--------------------------------------------------------------------*/

  node = dv->fe_space->mesh->node[CENTER];
  n0   = dv->fe_space->admin->n0_dof[CENTER];
  for (j = 0; j < 2; j++)
  {
    for (n = 0; n < DIM_OF_WORLD; n++)
      rvec[ibas][n] = v[dof[node][n0+j]][n];
    ibas++;
  }

  return((const REAL_D *) rvec);
}

/*--------------------------------------------------------------------*/
/*--- function for accessing a local DOF_SCHAR_VEC                 ---*/
/*--------------------------------------------------------------------*/

static const S_CHAR *get_schar_vec3(const EL *el, const DOF_SCHAR_VEC *dv,
                                    S_CHAR *vec)
{
  FUNCNAME("get_schar_vec3");
  static S_CHAR  my_vec[4];
  S_CHAR         *rvec = vec ? vec : my_vec;
  S_CHAR         *v = dv ? dv->vec : nil;
  int            ibas = 0, i, j, n0, node;
  DOF            **dof = el->dof;

  TEST_EXIT(v)("no DOF_SCHAR_VEC dv, or no dv->vec\n");

/*--------------------------------------------------------------------*/
/*--- DOFs at vertices                                             ---*/
/*--------------------------------------------------------------------*/

  node = dv->fe_space->mesh->node[VERTEX];
  n0   = dv->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[ibas++] = v[dof[node+i][n0]];

/*--------------------------------------------------------------------*/
/*--- DOFs at center                                               ---*/
/*--------------------------------------------------------------------*/

  node = dv->fe_space->mesh->node[CENTER];
  n0   = dv->fe_space->admin->n0_dof[CENTER];
  for (j = 0; j < 2; j++)
    rvec[ibas++] = v[dof[node][n0+j]];

  return(rvec);
}

/*--------------------------------------------------------------------*/
/*--- function for accessing a local DOF_UCHAR_VEC                 ---*/
/*--------------------------------------------------------------------*/

static const U_CHAR *get_uchar_vec3(const EL *el, const DOF_UCHAR_VEC *dv,
                                    U_CHAR *vec)
{
  FUNCNAME("get_uchar_vec3");
  static U_CHAR  my_vec[4];
  U_CHAR         *rvec = vec ? vec : my_vec;
  U_CHAR         *v = dv ? dv->vec : nil;
  int            ibas = 0, i, j, n0, node;
  DOF            **dof = el->dof;

  TEST_EXIT(v)("no DOF_UCHAR_VEC dv, or no dv->vec\n");

/*--------------------------------------------------------------------*/
/*--- DOFs at vertices                                             ---*/
/*--------------------------------------------------------------------*/

  node = dv->fe_space->mesh->node[VERTEX];
  n0   = dv->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[ibas++] = v[dof[node+i][n0]];

/*--------------------------------------------------------------------*/
/*--- DOFs at center                                               ---*/
/*--------------------------------------------------------------------*/

  node = dv->fe_space->mesh->node[CENTER];
  n0   = dv->fe_space->admin->n0_dof[CENTER];
  for (j = 0; j < 2; j++)
    rvec[ibas++] = v[dof[node][n0+j]];

  return(rvec);
}

/*--------------------------------------------------------------------*/
/*--- function for interpolaton of DOF_REAL_VECs during refinement ---*/
/*--------------------------------------------------------------------*/

static void refine_inter3(DOF_REAL_VEC *dv, RC_LIST_EL *list, int n_el)
{
  EL              *el;
  const DOF       *cdof;
  const REAL      *pvec;
  const BAS_FCTS  *bas_fcts = dv->fe_space->bas_fcts;
  const DOF_ADMIN *admin = dv->fe_space->admin;
  REAL            *v = dv->vec;

  el = list->el;
  pvec = (*bas_fcts->get_real_vec)(el, dv, nil);

/*--------------------------------------------------------------------*/
/*--- DOFs on first element, first child                           ---*/
/*--------------------------------------------------------------------*/

  cdof = (*bas_fcts->get_dof_indices)(el->child[0], admin, nil);

  v[cdof[1]] = (-0.0625*pvec[0] - 0.0625*pvec[1] + 0.5625*pvec[2] + 0.5625*pvec[3]);
  v[cdof[2]] = (0.3125*pvec[0] + 0.0625*pvec[1] + 0.9375*pvec[2] - 0.3125*pvec[3]);
  v[cdof[3]] = (pvec[2]);

/*--------------------------------------------------------------------*/
/*--- DOFs on first element, second child                          ---*/
/*--------------------------------------------------------------------*/

  cdof = (*bas_fcts->get_dof_indices)(el->child[1], admin, nil);

  v[cdof[2]] = (pvec[3]);
  v[cdof[3]] = (0.0625*pvec[0] + 0.3125*pvec[1] - 0.3125*pvec[2] + 0.9375*pvec[3]);

  return;
}

/*--------------------------------------------------------------------*/
/*--- function for interpolaton of DOF_REAL_D_VECs during refinement -*/
/*--------------------------------------------------------------------*/

static void refine_inter_d3(DOF_REAL_D_VEC *dv, RC_LIST_EL *list, int n_el)
{
  EL              *el;
  const DOF       *cdof;
  const REAL_D    *pvec;
  int             n;
  const BAS_FCTS  *bas_fcts = dv->fe_space->bas_fcts;
  const DOF_ADMIN *admin = dv->fe_space->admin;
  REAL_D          *v = dv->vec;

  el = list->el;
  pvec = (*bas_fcts->get_real_d_vec)(el, dv, nil);

/*--------------------------------------------------------------------*/
/*--- DOFs on first element, first child                           ---*/
/*--------------------------------------------------------------------*/

  cdof = (*bas_fcts->get_dof_indices)(el->child[0], admin, nil);

  for (n = 0; n < DIM_OF_WORLD; n++)
  {
    v[cdof[1]][n] = (-0.0625*pvec[0][n] - 0.0625*pvec[1][n] + 0.5625*pvec[2][n] + 0.5625*pvec[3][n]);
    v[cdof[2]][n] = (0.3125*pvec[0][n] + 0.0625*pvec[1][n] + 0.9375*pvec[2][n] - 0.3125*pvec[3][n]);
    v[cdof[3]][n] = (pvec[2][n]);
  }

/*--------------------------------------------------------------------*/
/*--- DOFs on first element, second child                          ---*/
/*--------------------------------------------------------------------*/

  cdof = (*bas_fcts->get_dof_indices)(el->child[1], admin, nil);

  for (n = 0; n < DIM_OF_WORLD; n++)
  {
    v[cdof[2]][n] = (pvec[3][n]);
    v[cdof[3]][n] = (0.0625*pvec[0][n] + 0.3125*pvec[1][n] - 0.3125*pvec[2][n] + 0.9375*pvec[3][n]);
  }

  return;
}

/*--------------------------------------------------------------------*/
/*--- function for interpolaton of DOF_REAL_VECs during coarsening ---*/
/*--------------------------------------------------------------------*/

static void coarse_inter3(DOF_REAL_VEC *dv, RC_LIST_EL *list, int n_el)
{
  EL              *el;
  const DOF       *pdof;
  const REAL      *cvec;
  const BAS_FCTS  *bas_fcts = dv->fe_space->bas_fcts;
  const DOF_ADMIN *admin = dv->fe_space->admin;
  REAL            *v = dv->vec;

  el = list->el;
  pdof = (*bas_fcts->get_dof_indices)(el, admin, nil);

/*--------------------------------------------------------------------*/
/*--- DOFs on first element, from first child                      ---*/
/*--------------------------------------------------------------------*/

  cvec = (*bas_fcts->get_real_vec)(el->child[0], dv, nil);

  v[pdof[2]] = (cvec[3]);

/*--------------------------------------------------------------------*/
/*--- DOFs on first element, second child                          ---*/
/*--------------------------------------------------------------------*/

  cvec = (*bas_fcts->get_real_vec)(el->child[1], dv, nil);

  v[pdof[3]] = (cvec[2]);

  return;
}

/*--------------------------------------------------------------------*/
/*--- function for interpolaton of DOF_REAL_D_VECs during coarsening -*/
/*--------------------------------------------------------------------*/

static void coarse_inter_d3(DOF_REAL_D_VEC *dv, RC_LIST_EL *list, int n_el)
{
  EL              *el;
  const DOF       *pdof;
  const REAL_D    *cvec;
  int             n;
  const BAS_FCTS  *bas_fcts = dv->fe_space->bas_fcts;
  const DOF_ADMIN *admin = dv->fe_space->admin;
  REAL_D          *v = dv->vec;

  el = list->el;
  pdof = (*bas_fcts->get_dof_indices)(el, admin, nil);

/*--------------------------------------------------------------------*/
/*--- DOFs on first element, from first child                      ---*/
/*--------------------------------------------------------------------*/

  cvec = (*bas_fcts->get_real_d_vec)(el->child[0], dv, nil);

  for (n = 0; n < DIM_OF_WORLD; n++)
  {
    v[pdof[2]][n] = (cvec[3][n]);
  }

/*--------------------------------------------------------------------*/
/*--- DOFs on first element, second child                          ---*/
/*--------------------------------------------------------------------*/

  cvec = (*bas_fcts->get_real_d_vec)(el->child[1], dv, nil);

  for (n = 0; n < DIM_OF_WORLD; n++)
  {
    v[pdof[3]][n] = (cvec[2][n]);
  }

  return;
}

/*--------------------------------------------------------------------*/
/*--- function for restriction of DOF_REAL_VECs during coarsening  ---*/
/*--------------------------------------------------------------------*/

static void coarse_restr3(DOF_REAL_VEC *dv, RC_LIST_EL *list, int n_el)
{
  EL              *el;
  const DOF       *pdof;
  const REAL      *cvec;
  const BAS_FCTS  *bas_fcts = dv->fe_space->bas_fcts;
  const DOF_ADMIN *admin = dv->fe_space->admin;
  REAL            *v = dv->vec;

  el = list->el;
  pdof = (*bas_fcts->get_dof_indices)(el, admin, nil);

/*--------------------------------------------------------------------*/
/*--- DOFs on first element, from first child                      ---*/
/*--------------------------------------------------------------------*/

  cvec = (*bas_fcts->get_real_vec)(el->child[0], dv, nil);

  v[pdof[0]] += (-0.0625*cvec[1] + 0.3125*cvec[2]);
  v[pdof[1]] += (-0.0625*cvec[1] + 0.0625*cvec[2]);
  v[pdof[2]]  = (0.5625*cvec[1] + 0.9375*cvec[2] + cvec[3]);
  v[pdof[3]]  = (0.5625*cvec[1] - 0.3125*cvec[2]);

/*--------------------------------------------------------------------*/
/*--- DOFs on first element, from second child                     ---*/
/*--------------------------------------------------------------------*/

  cvec = (*bas_fcts->get_real_vec)(el->child[1], dv, nil);

  v[pdof[0]] += (0.0625*cvec[3]);
  v[pdof[1]] += (0.3125*cvec[3]);
  v[pdof[2]] += (-0.3125*cvec[3]);
  v[pdof[3]] += (cvec[2] + 0.9375*cvec[3]);

  return;
}

/*--------------------------------------------------------------------*/
/*--- function for restriction of DOF_REAL_D_VECs during coarsening --*/
/*--------------------------------------------------------------------*/

static void coarse_restr_d3(DOF_REAL_D_VEC *dv, RC_LIST_EL *list, int n_el)
{
  EL              *el;
  const DOF       *pdof;
  const REAL_D    *cvec;
  int             n;
  const BAS_FCTS  *bas_fcts = dv->fe_space->bas_fcts;
  const DOF_ADMIN *admin = dv->fe_space->admin;
  REAL_D          *v = dv->vec;

  el = list->el;
  pdof = (*bas_fcts->get_dof_indices)(el, admin, nil);

/*--------------------------------------------------------------------*/
/*--- DOFs on first element, from first child                      ---*/
/*--------------------------------------------------------------------*/

  cvec = (*bas_fcts->get_real_d_vec)(el->child[0], dv, nil);

  for (n = 0; n < DIM_OF_WORLD; n++)
  {
    v[pdof[0]][n] += (-0.0625*cvec[1][n] + 0.3125*cvec[2][n]);
    v[pdof[1]][n] += (-0.0625*cvec[1][n] + 0.0625*cvec[2][n]);
    v[pdof[2]][n]  = (0.5625*cvec[1][n] + 0.9375*cvec[2][n] + cvec[3][n]);
    v[pdof[3]][n]  = (0.5625*cvec[1][n] - 0.3125*cvec[2][n]);
  }

/*--------------------------------------------------------------------*/
/*--- DOFs on first element, from second child                     ---*/
/*--------------------------------------------------------------------*/

  cvec = (*bas_fcts->get_real_d_vec)(el->child[1], dv, nil);

  for (n = 0; n < DIM_OF_WORLD; n++)
  {
    v[pdof[0]][n] += (0.0625*cvec[3][n]);
    v[pdof[1]][n] += (0.3125*cvec[3][n]);
    v[pdof[2]][n] += (-0.3125*cvec[3][n]);
    v[pdof[3]][n] += (cvec[2][n] + 0.9375*cvec[3][n]);
  }

  return;
}

/*--------------------------------------------------------------------*/
/*--- Collect all information about basis functions                ---*/
/*--------------------------------------------------------------------*/

static BAS_FCT *phi3[4] =
{
  phi3_0, phi3_1, phi3_2, phi3_3
};

static GRD_BAS_FCT *grd_phi3[4] =
{
  grd_phi3_0, grd_phi3_1, grd_phi3_2, grd_phi3_3
};
static D2_BAS_FCT *D2_phi3[4] =
{
  D2_phi3_0, D2_phi3_1, D2_phi3_2, D2_phi3_3
};

static BAS_FCTS lagrange3 =
{
  "lagrange3", 4, 3, {1, 2},
  nil, phi3, grd_phi3, D2_phi3,
  get_dof_indices3,
  get_bound3,
  interpol3,
  interpol_d3,
  get_int_vec3,
  get_real_vec3,
  get_real_d_vec3,
  get_uchar_vec3,
  get_schar_vec3,
  refine_inter3,
  coarse_inter3,
  coarse_restr3,
  refine_inter_d3,
  coarse_inter_d3,
  coarse_restr_d3,
  (void *) coords3
};
