/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     lagrange_2.c                                                   */
/*                                                                          */
/* description:  piecewise quadratic Lagrange elements in 2d                */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#define N_BAS2  10

static const REAL   bary2[N_BAS2][DIM+1] = {{1.0, 0.0, 0.0, 0.0},
					 {0.0, 1.0, 0.0, 0.0},
					 {0.0, 0.0, 1.0, 0.0},
					 {0.0, 0.0, 0.0, 1.0},
					 {0.5, 0.5, 0.0, 0.0},
					 {0.5, 0.0, 0.5, 0.0},
					 {0.5, 0.0, 0.0, 0.5},
					 {0.0, 0.5, 0.5, 0.0},
					 {0.0, 0.5, 0.0, 0.5},
					 {0.0, 0.0, 0.5, 0.5}};
#if 0
static const int    fct_at_node2[N_BAS2] = {0,1,2,3,4,5,6,7,8,9};

static const int    n_dof2[DIM+1] = {1, 1, 0, 0};
#endif

/*--------------------------------------------------------------------------*/
/*  basisfunction at vertex 0                                               */
/*--------------------------------------------------------------------------*/

static REAL phi2v0(const REAL lambda[DIM+1])
{
  return(lambda[0]*(2.0*lambda[0] - 1.0));
}

static const REAL *grd_phi2v0(const REAL lambda[DIM+1])
{
  static REAL  grd[DIM+1] = {0.0, 0.0, 0.0, 0.0};
  grd[0] = 4.0*lambda[0] - 1.0;
  return((const REAL *) grd);
}

static const REAL (*D2_phi2v0(const REAL *lambda))[DIM+1]
{
  static const REAL D2[DIM+1][DIM+1] = {{4, 0, 0, 0}, {0, 0, 0, 0}, 
					{0, 0, 0, 0}, {0, 0, 0, 0}};
  return(D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction at vertex 1                                               */
/*--------------------------------------------------------------------------*/

static REAL phi2v1(const REAL lambda[DIM+1])
{
  return(lambda[1]*(2.0*lambda[1] - 1.0));
}

static const REAL *grd_phi2v1(const REAL lambda[DIM+1])
{
  static REAL  grd[DIM+1] = {0.0, 0.0, 0.0, 0.0};
  grd[1] = 4.0*lambda[1] - 1.0;
  return((const REAL *) grd);
}

static const REAL (*D2_phi2v1(const REAL *lambda))[DIM+1]
{
  static const REAL D2[DIM+1][DIM+1] = {{0, 0, 0, 0}, {0, 4, 0, 0}, 
					{0, 0, 0, 0}, {0, 0, 0, 0}};
  return(D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction at vertex 2                                               */
/*--------------------------------------------------------------------------*/

static REAL phi2v2(const REAL lambda[DIM+1])
{
  return(lambda[2]*(2.0*lambda[2] - 1.0));
}

static const REAL *grd_phi2v2(const REAL lambda[DIM+1])
{
  static REAL  grd[DIM+1] = {0.0, 0.0, 0.0, 0.0};
  grd[2] = 4.0*lambda[2] - 1.0;
  return((const REAL *) grd);
}

static const REAL (*D2_phi2v2(const REAL *lambda))[DIM+1]
{
  static const REAL D2[DIM+1][DIM+1] = {{0, 0, 0, 0}, {0, 0, 0, 0}, 
					{0, 0, 4, 0}, {0, 0, 0, 0}};
  return(D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction at vertex 3                                               */
/*--------------------------------------------------------------------------*/

static REAL phi2v3(const REAL lambda[DIM+1])
{
  return(lambda[3]*(2.0*lambda[3] - 1.0));
}

static const REAL *grd_phi2v3(const REAL lambda[DIM+1])
{
  static REAL  grd[DIM+1] = {0.0, 0.0, 0.0, 0.0};
  grd[3] = 4.0*lambda[3] - 1.0;
  return((const REAL *) grd);
}

static const REAL (*D2_phi2v3(const REAL *lambda))[DIM+1]
{
  static const REAL D2[DIM+1][DIM+1] = {{0, 0, 0, 0}, {0, 0, 0, 0}, 
					{0, 0, 0, 0}, {0, 0, 0, 4}};
  return(D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction at edge 0                                                 */
/*--------------------------------------------------------------------------*/

static REAL phi2e0(const REAL lambda[DIM+1])
{
  return(4.0*lambda[0]*lambda[1]);
}

static const REAL *grd_phi2e0(const REAL lambda[DIM+1])
{
  static REAL  grd[DIM+1] = {0.0, 0.0, 0.0, 0.0};
  grd[0] = 4.0*lambda[1];
  grd[1] = 4.0*lambda[0];
  return((const REAL *) grd);
}

static const REAL (*D2_phi2e0(const REAL *lambda))[DIM+1]
{
  static const REAL D2[DIM+1][DIM+1] = {{0, 4, 0, 0}, {4, 0, 0, 0}, 
					{0, 0, 0, 0}, {0, 0, 0, 0}};
  return(D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction at edge 1                                                 */
/*--------------------------------------------------------------------------*/

static REAL phi2e1(const REAL lambda[DIM+1])
{
  return(4.0*lambda[0]*lambda[2]);
}

static const REAL *grd_phi2e1(const REAL lambda[DIM+1])
{
  static REAL  grd[DIM+1] = {0.0, 0.0, 0.0, 0.0};
  grd[0] = 4.0*lambda[2];
  grd[2] = 4.0*lambda[0];
  return((const REAL *) grd);
}

static const REAL (*D2_phi2e1(const REAL *lambda))[DIM+1]
{
  static const REAL D2[DIM+1][DIM+1] = {{0, 0, 4, 0}, {0, 0, 0, 0}, 
					{4, 0, 0, 0}, {0, 0, 0, 0}};
  return(D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction at edge 2                                                 */
/*--------------------------------------------------------------------------*/

static REAL phi2e2(const REAL lambda[DIM+1])
{
  return(4.0*lambda[0]*lambda[3]);
}

static const REAL *grd_phi2e2(const REAL lambda[DIM+1])
{
  static REAL  grd[DIM+1] = {0.0, 0.0, 0.0, 0.0};
  grd[0] = 4.0*lambda[3];
  grd[3] = 4.0*lambda[0];
  return((const REAL *) grd);
}

static const REAL (*D2_phi2e2(const REAL *lambda))[DIM+1]
{
  static const REAL D2[DIM+1][DIM+1] = {{0, 0, 0, 4}, {0, 0, 0, 0}, 
					{0, 0, 0, 0}, {4, 0, 0, 0}};
  return(D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction at edge 3                                                 */
/*--------------------------------------------------------------------------*/

static REAL phi2e3(const REAL lambda[DIM+1])
{
  return(4.0*lambda[1]*lambda[2]);
}

static const REAL *grd_phi2e3(const REAL lambda[DIM+1])
{
  static REAL  grd[DIM+1] = {0.0, 0.0, 0.0, 0.0};
  grd[1] = 4.0*lambda[2];
  grd[2] = 4.0*lambda[1];
  return((const REAL *) grd);
}

static const REAL (*D2_phi2e3(const REAL *lambda))[DIM+1]
{
  static const REAL D2[DIM+1][DIM+1] = {{0, 0, 0, 0}, {0, 0, 4, 0}, 
					{0, 4, 0, 0}, {0, 0, 0, 0}};
  return(D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction at edge 4                                                 */
/*--------------------------------------------------------------------------*/

static REAL phi2e4(const REAL lambda[DIM+1])
{
  return(4.0*lambda[1]*lambda[3]);
}

static const REAL *grd_phi2e4(const REAL lambda[DIM+1])
{
  static REAL  grd[DIM+1] = {0.0, 0.0, 0.0, 0.0};
  grd[1] = 4.0*lambda[3];
  grd[3] = 4.0*lambda[1];
  return((const REAL *) grd);
}

static const REAL (*D2_phi2e4(const REAL *lambda))[DIM+1]
{
  static const REAL D2[DIM+1][DIM+1] = {{0, 0, 0, 0}, {0, 0, 0, 4}, 
					{0, 0, 0, 0}, {0, 4, 0, 0}};
  return(D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction at edge 5                                                 */
/*--------------------------------------------------------------------------*/

static REAL phi2e5(const REAL lambda[DIM+1])
{
  return(4.0*lambda[2]*lambda[3]);
}

static const REAL *grd_phi2e5(const REAL lambda[DIM+1])
{
  static REAL  grd[DIM+1] = {0.0, 0.0, 0.0, 0.0};
  grd[2] = 4.0*lambda[3];
  grd[3] = 4.0*lambda[2];
  return((const REAL *) grd);
}

static const REAL (*D2_phi2e5(const REAL *lambda))[DIM+1]
{
  static const REAL D2[DIM+1][DIM+1] = {{0, 0, 0, 0}, {0, 0, 0, 0}, 
					{0, 0, 0, 4}, {0, 0, 4, 0}};
  return(D2);
}

/*--------------------------------------------------------------------------*/
/*  functions for combining basisfunctions with coefficients                */
/*--------------------------------------------------------------------------*/

static const DOF *get_dof_indices2(const EL *el, const DOF_ADMIN *admin,
				   DOF *idof)
{
  static DOF  index_vec[N_VERTICES+N_EDGES];
  int         i, n0;
  DOF         *rvec = idof ? idof : index_vec;
  DOF         **dof = el->dof;
  n0 = admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[i] = dof[i][n0];
  n0 = admin->n0_dof[EDGE];
  for (i = N_VERTICES; i < N_VERTICES+N_EDGES; i++)
    rvec[i] = dof[i][n0];

  return((const DOF *) rvec);
}

static const S_CHAR *get_bound2(const EL_INFO *el_info, S_CHAR *bound)
{
  FUNCNAME("get_bound2");
  static S_CHAR  bound_vec[N_VERTICES+N_EDGES];
  S_CHAR         *rvec = bound ? bound : bound_vec;
  int            i;

  TEST_FLAG(FILL_BOUND, el_info);

  for (i = 0; i < N_VERTICES; i++)
    rvec[i] = el_info->bound[i];
  for (i = 0; i < N_EDGES; i++)
    rvec[N_VERTICES+i] = GET_BOUND(el_info->boundary[N_FACES+i]);

  return((const S_CHAR *) rvec);
}

static const int *get_int_vec2(const EL *el, const DOF_INT_VEC *vec,
				int *ivec)
{
  FUNCNAME("get_int_vec2");
  int           i, n0;
  static int    local_vec[N_VERTICES+N_EDGES];
  int           *v = nil, *rvec = ivec ? ivec : local_vec;
  DOF           **dof = el->dof;

  GET_DOF_VEC(v, vec);

  n0 = vec->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[i] = v[dof[i][n0]];
  n0 = vec->fe_space->admin->n0_dof[EDGE];
  for (i = N_VERTICES; i < N_VERTICES+N_EDGES; i++)
    rvec[i] = v[dof[i][n0]];

  return((const int *) rvec);
}

static const REAL *get_real_vec2(const EL *el, const DOF_REAL_VEC *vec,
				 REAL *Rvec)
{
  FUNCNAME("get_real_vec2");
  int            i, n0;
  static REAL    local_vec[N_VERTICES+N_EDGES];
  REAL           *v = nil, *rvec = Rvec ? Rvec : local_vec;
  DOF            **dof = el->dof;

  GET_DOF_VEC(v, vec);

  n0 = vec->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[i] = v[dof[i][n0]];
  n0 = vec->fe_space->admin->n0_dof[EDGE];
  for (i = N_VERTICES; i < N_VERTICES+N_EDGES; i++)
    rvec[i] = v[dof[i][n0]];

  return((const REAL *) rvec);
}

static const REAL_D *get_real_d_vec2(const EL *el, const DOF_REAL_D_VEC *vec,
				     REAL_D *Rvec)
{
  FUNCNAME("get_real_d_vec2");
  int            i, k, n0;
  static REAL_D  local_vec[N_BAS2];
  REAL_D         *v = nil, *rvec = Rvec ? Rvec : local_vec;
  DOF            **dof = el->dof;

  GET_DOF_VEC(v, vec);

  n0 = vec->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    for (k = 0; k < DIM_OF_WORLD; k++)
      rvec[i][k] = v[dof[i][n0]][k];
  n0 = vec->fe_space->admin->n0_dof[EDGE];
  for (i = N_VERTICES; i < N_VERTICES+N_EDGES; i++)
    for (k = 0; k < DIM_OF_WORLD; k++)
      rvec[i][k] = v[dof[i][n0]][k];

  return((const REAL_D *) rvec);
}

static const U_CHAR *get_uchar_vec2(const EL *el, const DOF_UCHAR_VEC *vec,
				     U_CHAR *uvec)
{
  FUNCNAME("get_uchar_vec2");
  int            i, n0;
  static U_CHAR  local_vec[N_VERTICES+N_EDGES];
  U_CHAR         *v = nil, *rvec = uvec ? uvec : local_vec;
  DOF            **dof = el->dof;

  GET_DOF_VEC(v, vec);

  n0 = vec->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[i] = v[dof[i][n0]];
  n0 = vec->fe_space->admin->n0_dof[EDGE];
  for (i = N_VERTICES; i < N_VERTICES+N_EDGES; i++)
    rvec[i] = v[dof[i][n0]];

  return((const U_CHAR *) rvec);
}

static const S_CHAR *get_schar_vec2(const EL *el, const DOF_SCHAR_VEC *vec,
				     S_CHAR *svec)
{
  FUNCNAME("get_schar_vec2");
  int            i, n0;
  static S_CHAR  local_vec[N_VERTICES+N_EDGES];
  S_CHAR         *v = nil, *rvec = svec ? svec : local_vec;
  DOF            **dof = el->dof;

  GET_DOF_VEC(v, vec);

  n0 = vec->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[i] = v[dof[i][n0]];
  n0 = vec->fe_space->admin->n0_dof[EDGE];
  for (i = N_VERTICES; i < N_VERTICES+N_EDGES; i++)
    rvec[i] = v[dof[i][n0]];

  return((const S_CHAR *) rvec);
}

static const REAL *interpol2(const EL_INFO *el_info, int no, 
			     const int *b_no, REAL (*f)(const REAL_D),
			     REAL *vec)
{
  FUNCNAME("interpol2");
  static REAL  inter[N_VERTICES+N_EDGES];
  REAL         *rvec = vec ? vec : inter;
  REAL_D       x;
  const REAL_D *coord;
  int          i, j, k, l;
  const PARAMETRIC *parametric = el_info->mesh->parametric;

  if (parametric) 
    ERROR_EXIT("Sorry, parametric support for lagrange2 not implemented yet!\n");

  TEST_FLAG(FILL_COORDS, el_info);
  coord = el_info->coord;

  if (b_no)
  {
    if (no <= 0  ||  no > N_VERTICES+N_EDGES)
    {
      MSG("something is wrong, doing nothing\n");
      rvec[0] = 0.0;
      return((const REAL *) rvec);
    }
  
    for (i = 0; i < no; i++)
    {
      if ((l = b_no[i] - N_VERTICES) < 0)
      {
	rvec[i] = f(coord[b_no[i]]);
      }
      else
      {
	k = vertex_of_edge[l][0];
	l = vertex_of_edge[l][1];
	for (j = 0; j < DIM_OF_WORLD; j++)
	  x[j] = (coord[k][j] + coord[l][j])*0.5;
	  rvec[i] = f(x);
      }
    }
  }
  else
  {
    for (i = 0; i < N_VERTICES; i++)
      rvec[i] = f(coord[i]);

    for (i = 0; i < N_EDGES; i++)
    {
      k = vertex_of_edge[i][0];
      l = vertex_of_edge[i][1];
      for (j = 0; j < DIM_OF_WORLD; j++)
	x[j] = (coord[k][j] + coord[l][j])*0.5;

      rvec[N_VERTICES+i] = f(x);
    }
  }

  return((const REAL *) rvec);
}

static const REAL_D *interpol_d2(const EL_INFO *el_info, int no, 
				 const int *b_no,
				 const REAL *(*f)(const REAL_D, REAL_D),
				 REAL_D *vec)
{
  FUNCNAME("interpol_d2");
  static REAL_D  inter[N_VERTICES+N_EDGES];
  REAL_D         *rvec = vec ? vec : inter;
  REAL_D         x;
  const REAL_D   *coord;
  int            i, j, k, l;
  const PARAMETRIC *parametric = el_info->mesh->parametric;

  if (parametric) 
    ERROR_EXIT("Sorry, parametric support for lagrange2 not implemented yet!\n");

  TEST_FLAG(FILL_COORDS, el_info);
  
  coord = el_info->coord;
  if (b_no)
  {
    if (no <= 0  ||  no > N_VERTICES+N_EDGES)
    {
      ERROR("something is wrong, doing nothing\n");
      for (i = 0; i < DIM_OF_WORLD; i++)
	rvec[0][i] = 0.0;
      return((const REAL_D *) rvec);
    }
    for (i = 0; i < no; i++)
    {
      if ((l = b_no[i] - N_VERTICES) < 0)
      {
	(*f)(coord[b_no[i]], rvec[i]);
      }
      else
      {
	k = vertex_of_edge[l][0];
	l = vertex_of_edge[l][1];
	
	for (j = 0; j < DIM_OF_WORLD; j++)
	{
	  x[j] = (coord[k][j] + coord[l][j])*0.5;
	}
	(*f)(x, rvec[i]);
      }
    }
  }
  else
  {
    for (i = 0; i < N_VERTICES; i++)
      (*f)(coord[i], rvec[i]);

    for (i = 0; i < N_EDGES; i++)
    {
      k = vertex_of_edge[i][0];
      l = vertex_of_edge[i][1];
      for (j = 0; j < DIM_OF_WORLD; j++)
      {
	x[j] = (coord[k][j] + coord[l][j])*0.5;
      }
      (*f)(x, rvec[N_VERTICES+i]);
    }
  }
  
  return((const REAL_D *) rvec);
}


/*--------------------------------------------------------------------------*/
/*  functions for interpolation/ restriction during refinement/coarsening   */
/*--------------------------------------------------------------------------*/

static void real_refine_inter2(DOF_REAL_VEC *drv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_refine_inter2");
  EL        *el;
  REAL      *v = nil;
  const DOF *cdof;
  DOF       pdof[N_BAS2], cdofi;
  int       i, lr_set;
  int       node0, n0;
  const DOF       *(*get_dof_indices)(const EL *, const DOF_ADMIN *, DOF *);
  const DOF_ADMIN *admin;

  if (n < 1) return;
  el = list->el;

  GET_DOF_VEC(v, drv);
  if (!drv->fe_space)
  {
    ERROR("no fe_space in dof_real_vec %s\n", NAME(drv));
    return;
  }
  else if (!drv->fe_space->bas_fcts)
  {
    ERROR("no basis functions in fe_space %s\n", NAME(drv->fe_space));
    return;
  }
  get_dof_indices = drv->fe_space->bas_fcts->get_dof_indices;
  GET_STRUCT(admin,drv->fe_space);

  get_dof_indices(el, admin, pdof);

  node0 = drv->fe_space->mesh->node[EDGE];
  n0 = admin->n0_dof[EDGE];

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

  cdof = get_dof_indices(el->child[0], admin, nil);

  v[cdof[3]] = (v[pdof[4]]);
  v[cdof[6]] = (0.375*v[pdof[0]] - 0.125*v[pdof[1]] 
		+ 0.75*v[pdof[4]]);
  v[cdof[8]] = (0.125*(-v[pdof[0]] - v[pdof[1]]) + 0.25*v[pdof[4]]
		+ 0.5*(v[pdof[5]] + v[pdof[7]]));
  v[cdof[9]] = (0.125*(-v[pdof[0]] - v[pdof[1]]) + 0.25*v[pdof[4]]
		+ 0.5*(v[pdof[6]] + v[pdof[8]]));

/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/
  
  cdofi = el->child[1]->dof[node0+2][n0];
  v[cdofi] = (-0.125*v[pdof[0]] + 0.375*v[pdof[1]] 
	      + 0.75*v[pdof[4]]);

/*--------------------------------------------------------------------------*/
/*   adjust neighbour values                                                */
/*--------------------------------------------------------------------------*/
  
  for (i = 1; i < n; i++)
  {
    el = list[i].el;
    get_dof_indices(el, admin, pdof);

    lr_set = 0;
    if (list[i].neigh[0]  &&  list[i].neigh[0]->no < i)
      lr_set = 1;

    if (list[i].neigh[1]  &&  list[i].neigh[1]->no < i)
      lr_set += 2;

    TEST_EXIT(lr_set)("no values set on both neighbours\n");

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

    switch (lr_set)
    {
    case 1:
      cdofi = el->child[0]->dof[node0+4][n0];      
      v[cdofi] = (0.125*(-v[pdof[0]] - v[pdof[1]]) + 0.25*v[pdof[4]]
		  + 0.5*(v[pdof[5]] + v[pdof[7]]));
      break;
    case 2:
      cdofi = el->child[0]->dof[node0+5][n0];      
      v[cdofi] = (0.125*(-v[pdof[0]] - v[pdof[1]]) + 0.25*v[pdof[4]]
		  + 0.5*(v[pdof[6]] + v[pdof[8]]));
    }
  }
  return;
}

static void real_coarse_inter2(DOF_REAL_VEC *drv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_coarse_inter2");
  EL        *el;
  REAL      *v = nil;
  int       cdof, pdof;
  const DOF_ADMIN *admin;
  MESH      *mesh = nil;

  if (n < 1) return;
  el = list->el;

  GET_DOF_VEC(v, drv);
  if (!drv->fe_space)
  {
    ERROR("no fe_space in dof_real_vec %s\n", NAME(drv));
    return;
  }
  else if (!drv->fe_space->bas_fcts)
  {
    ERROR("no basis functions in fe_space %s\n", NAME(drv->fe_space));
    return;
  }
  GET_STRUCT(admin,drv->fe_space);
  GET_STRUCT(mesh,drv->fe_space);

  cdof = el->child[0]->dof[mesh->node[VERTEX]+3][admin->n0_dof[VERTEX]]; 
  pdof = el->dof[mesh->node[EDGE]][admin->n0_dof[EDGE]]; 
  v[pdof] = v[cdof];

  return;
}

static void real_coarse_restr2(DOF_REAL_VEC *drv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_coarse_restr2");
  EL        *el;
  REAL      *v = nil;
  const DOF *cdof;
  DOF       pdof[N_BAS2], cdofi;
  int       i, lr_set;
  int       node0, n0;
  const DOF       *(*get_dof_indices)(const EL *, const DOF_ADMIN *, DOF *);
  const DOF_ADMIN *admin;

  if (n < 1) return;
  el = list->el;

  GET_DOF_VEC(v, drv);
  if (!drv->fe_space)
  {
    ERROR("no fe_space in dof_real_vec %s\n", NAME(drv));
    return;
  }
  else if (!drv->fe_space->bas_fcts)
  {
    ERROR("no basis functions in fe_space %s\n", NAME(drv->fe_space));
    return;
  }
  get_dof_indices = drv->fe_space->bas_fcts->get_dof_indices;
  GET_STRUCT(admin,drv->fe_space);

  get_dof_indices(el, admin, pdof);

  node0 = drv->fe_space->mesh->node[EDGE];
  n0 = admin->n0_dof[EDGE];

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

  cdof = get_dof_indices(el->child[0], admin, nil);

  v[pdof[0]] += (0.375*v[cdof[6]] + 0.125*(-v[cdof[8]] - v[cdof[9]]));
  v[pdof[1]] += (0.125*(-v[cdof[6]] - v[cdof[8]] - v[cdof[9]]));
  v[pdof[4]] = (v[cdof[3]] + 0.75*v[cdof[6]] 
		+ 0.25*(v[cdof[8]] + v[cdof[9]]));
  v[pdof[5]] += (0.5*v[cdof[8]]);
  v[pdof[6]] += (0.5*v[cdof[9]]);
  v[pdof[7]] += (0.5*v[cdof[8]]);
  v[pdof[8]] += (0.5*v[cdof[9]]);

/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/

  cdof = get_dof_indices(el->child[1], admin, nil);
  cdofi = el->child[1]->dof[node0+2][n0];

  v[pdof[0]] += (-0.125*v[cdofi]);
  v[pdof[1]] += (0.375*v[cdofi]);
  v[pdof[4]] += (0.75*v[cdofi]);

/*--------------------------------------------------------------------------*/
/*   adjust neighbour values                                                */
/*--------------------------------------------------------------------------*/
  
  for (i = 1; i < n; i++)
  {
    el = list[i].el;
    get_dof_indices(el, admin, pdof);

    lr_set = 0;
    if (list[i].neigh[0]  &&  list[i].neigh[0]->no < i)
      lr_set = 1;

    if (list[i].neigh[1]  &&  list[i].neigh[1]->no < i)
      lr_set += 2;

    TEST_EXIT(lr_set)("no values set on both neighbours\n");

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

    cdof = get_dof_indices(el->child[0], admin, nil);

    switch(lr_set)
    {
    case 1:
      cdofi = el->child[0]->dof[node0+4][n0];      
      v[pdof[0]] += (-0.125*v[cdofi]);
      v[pdof[1]] += (-0.125*v[cdofi]);
      v[pdof[4]] += (0.25*v[cdofi]);
      v[pdof[5]] += (0.5*v[cdofi]);
      v[pdof[7]] += (0.5*v[cdofi]);
      break;
    case 2:
      cdofi = el->child[0]->dof[node0+5][n0];      
      v[pdof[0]] += (-0.125*v[cdofi]);
      v[pdof[1]] += (-0.125*v[cdofi]);
      v[pdof[4]] += (0.25*v[cdofi]);
      v[pdof[6]] += (0.5*v[cdofi]);
      v[pdof[8]] += (0.5*v[cdofi]);
      break;
    }
  }
  return;
}

static void real_d_refine_inter2(DOF_REAL_D_VEC *drdv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_d_refine_inter2");
  EL        *el;
  REAL_D    *v = nil;
  const DOF *cd;
  DOF       pd[N_BAS2], cdi;
  int       i, k, lr_set;
  int       node0, n0;
  const DOF       *(*get_dof_indices)(const EL *, const DOF_ADMIN *, DOF *);
  const DOF_ADMIN *admin;

  if (n < 1) return;
  el = list->el;

  GET_DOF_VEC(v, drdv);
  if (!drdv->fe_space)
  {
    ERROR("no fe_space in dof_real_d_vec %s\n", NAME(drdv));
    return;
  }
  else if (!drdv->fe_space->bas_fcts)
  {
    ERROR("no basis functions in fe_space %s\n", NAME(drdv->fe_space));
    return;
  }
  get_dof_indices = drdv->fe_space->bas_fcts->get_dof_indices;
  GET_STRUCT(admin,drdv->fe_space);

  get_dof_indices(el, admin, pd);

  node0 = drdv->fe_space->mesh->node[EDGE];
  n0 = admin->n0_dof[EDGE];

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

  cd = get_dof_indices(el->child[0], admin, nil);

  for (k = 0; k < DIM_OF_WORLD; k++)
  {
    v[cd[3]][k] = (v[pd[4]][k]);
    v[cd[6]][k] = (0.375*v[pd[0]][k] - 0.125*v[pd[1]][k] + 0.75*v[pd[4]][k]);
    v[cd[8]][k] = (0.125*(-v[pd[0]][k] - v[pd[1]][k]) + 0.25*v[pd[4]][k]
		   + 0.5*(v[pd[5]][k] + v[pd[7]][k]));
    v[cd[9]][k] = (0.125*(-v[pd[0]][k] - v[pd[1]][k]) + 0.25*v[pd[4]][k]
		   + 0.5*(v[pd[6]][k] + v[pd[8]][k]));
  }
/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/
  
  cdi = el->child[1]->dof[node0+2][n0];
  for (k = 0; k < DIM_OF_WORLD; k++)
    v[cdi][k] = (-0.125*v[pd[0]][k] + 0.375*v[pd[1]][k] + 0.75*v[pd[4]][k]);

/*--------------------------------------------------------------------------*/
/*   adjust neighbour values                                                */
/*--------------------------------------------------------------------------*/
  
  for (i = 1; i < n; i++)
  {
    el = list[i].el;
    get_dof_indices(el, admin, pd);

    lr_set = 0;
    if (list[i].neigh[0]  &&  list[i].neigh[0]->no < i)
      lr_set = 1;

    if (list[i].neigh[1]  &&  list[i].neigh[1]->no < i)
      lr_set += 2;

    TEST_EXIT(lr_set)("no values set on both neighbours\n");

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

    switch (lr_set)
    {
    case 1:
      cdi = el->child[0]->dof[node0+4][n0];      
      for (k = 0; k < DIM_OF_WORLD; k++)
	v[cdi][k] = (0.125*(-v[pd[0]][k] - v[pd[1]][k]) + 0.25*v[pd[4]][k]
		     + 0.5*(v[pd[5]][k] + v[pd[7]][k]));
      break;
    case 2:
      cdi = el->child[0]->dof[node0+5][n0];      
      for (k = 0; k < DIM_OF_WORLD; k++)
	v[cdi][k] = (0.125*(-v[pd[0]][k] - v[pd[1]][k]) + 0.25*v[pd[4]][k]
		     + 0.5*(v[pd[6]][k] + v[pd[8]][k]));
    }
  }
  return;
}

static void real_d_coarse_inter2(DOF_REAL_D_VEC *drdv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_d_coarse_inter2");
  EL        *el;
  REAL_D    *v = nil;
  int       cd, pd, k;
  const DOF_ADMIN *admin;
  MESH      *mesh = nil;

  if (n < 1) return;
  el = list->el;

  GET_DOF_VEC(v, drdv);
  if (!drdv->fe_space)
  {
    ERROR("no fe_space in dof_real_d_vec %s\n", NAME(drdv));
    return;
  }
  else if (!drdv->fe_space->bas_fcts)
  {
    ERROR("no basis functions in fe_space %s\n", NAME(drdv->fe_space));
    return;
  }
  GET_STRUCT(admin,drdv->fe_space);
  GET_STRUCT(mesh,drdv->fe_space);

  cd = el->child[0]->dof[mesh->node[VERTEX]+3][admin->n0_dof[VERTEX]]; 
  pd = el->dof[mesh->node[EDGE]][admin->n0_dof[EDGE]]; 
  for (k = 0; k < DIM_OF_WORLD; k++)
    v[pd][k] = v[cd][k];

  return;
}

static void real_d_coarse_restr2(DOF_REAL_D_VEC *drdv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_d_coarse_restr2");
  EL        *el;
  REAL_D    *v = nil;
  const DOF *cd;
  DOF       pd[N_BAS2], cdi;
  int       i, k, lr_set;
  int       node0, n0;
  const DOF       *(*get_dof_indices)(const EL *, const DOF_ADMIN *, DOF *);
  const DOF_ADMIN *admin;

  if (n < 1) return;
  el = list->el;

  GET_DOF_VEC(v, drdv);
  if (!drdv->fe_space)
  {
    ERROR("no fe_space in dof_real_d_vec %s\n", NAME(drdv));
    return;
  }
  else if (!drdv->fe_space->bas_fcts)
  {
    ERROR("no basis functions in fe_space %s\n", NAME(drdv->fe_space));
    return;
  }
  get_dof_indices = drdv->fe_space->bas_fcts->get_dof_indices;
  GET_STRUCT(admin,drdv->fe_space);

  get_dof_indices(el, admin, pd);

  node0 = drdv->fe_space->mesh->node[EDGE];
  n0 = admin->n0_dof[EDGE];

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

  cd = get_dof_indices(el->child[0], admin, nil);

  for (k = 0; k < DIM_OF_WORLD; k++)
  {
    v[pd[0]][k] += (0.375*v[cd[6]][k] + 0.125*(-v[cd[8]][k] - v[cd[9]][k]));
    v[pd[1]][k] += (0.125*(-v[cd[6]][k] - v[cd[8]][k] - v[cd[9]][k]));
    v[pd[4]][k] = (v[cd[3]][k] + 0.75*v[cd[6]][k] 
		   + 0.25*(v[cd[8]][k] + v[cd[9]][k]));
    v[pd[5]][k] += (0.5*v[cd[8]][k]);
    v[pd[6]][k] += (0.5*v[cd[9]][k]);
    v[pd[7]][k] += (0.5*v[cd[8]][k]);
    v[pd[8]][k] += (0.5*v[cd[9]][k]);
  }
/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/

  cd = get_dof_indices(el->child[1], admin, nil);
  cdi = el->child[1]->dof[node0+2][n0];

  for (k = 0; k < DIM_OF_WORLD; k++)
  {
    v[pd[0]][k] += (-0.125*v[cdi][k]);
    v[pd[1]][k] += (0.375*v[cdi][k]);
    v[pd[4]][k] += (0.75*v[cdi][k]);
  }

/*--------------------------------------------------------------------------*/
/*   adjust neighbour values                                                */
/*--------------------------------------------------------------------------*/
  
  for (i = 1; i < n; i++)
  {
    el = list[i].el;
    get_dof_indices(el, admin, pd);

    lr_set = 0;
    if (list[i].neigh[0]  &&  list[i].neigh[0]->no < i)
      lr_set = 1;

    if (list[i].neigh[1]  &&  list[i].neigh[1]->no < i)
      lr_set += 2;

    TEST_EXIT(lr_set)("no values set on both neighbours\n");

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

    cd = get_dof_indices(el->child[0], admin, nil);

    switch(lr_set)
    {
    case 1:
      cdi = el->child[0]->dof[node0+4][n0];      
      for (k = 0; k < DIM_OF_WORLD; k++)
      {
	v[pd[0]][k] += (-0.125*v[cdi][k]);
	v[pd[1]][k] += (-0.125*v[cdi][k]);
	v[pd[4]][k] += (0.25*v[cdi][k]);
	v[pd[5]][k] += (0.5*v[cdi][k]);
	v[pd[7]][k] += (0.5*v[cdi][k]);
      }
      break;
    case 2:
      cdi = el->child[0]->dof[node0+5][n0];      
      for (k = 0; k < DIM_OF_WORLD; k++)
      {
	v[pd[0]][k] += (-0.125*v[cdi][k]);
	v[pd[1]][k] += (-0.125*v[cdi][k]);
	v[pd[4]][k] += (0.25*v[cdi][k]);
	v[pd[6]][k] += (0.5*v[cdi][k]);
	v[pd[8]][k] += (0.5*v[cdi][k]);
      }
      break;
    }
  }
  return;
}

static BAS_FCT      *phi2[N_BAS2]     = {phi2v0, phi2v1,
					 phi2v2, phi2v3,
					 phi2e0, phi2e1,
					 phi2e2, phi2e3,
					 phi2e4, phi2e5};
static GRD_BAS_FCT  *grd_phi2[N_BAS2] = {grd_phi2v0, grd_phi2v1,
					 grd_phi2v2, grd_phi2v3, 
					 grd_phi2e0, grd_phi2e1,
					 grd_phi2e2, grd_phi2e3,
					 grd_phi2e4, grd_phi2e5};
static D2_BAS_FCT   *D2_phi2[N_BAS2]  = {D2_phi2v0, D2_phi2v1,
					 D2_phi2v2, D2_phi2v3,
					 D2_phi2e0, D2_phi2e1,
					 D2_phi2e2, D2_phi2e3,
					 D2_phi2e4, D2_phi2e5};

static const BAS_FCTS  lagrange2 = {"lagrange2", N_BAS2, 2, {1, 1, 0, 0},
				    nil, phi2, grd_phi2, D2_phi2, 
				    get_dof_indices2, 
				    get_bound2, 
				    interpol2,
				    interpol_d2,
				    get_int_vec2,
				    get_real_vec2,
				    get_real_d_vec2,
				    get_uchar_vec2,
				    get_schar_vec2,
				    real_refine_inter2,
				    real_coarse_inter2,
				    real_coarse_restr2,
				    real_d_refine_inter2,
				    real_d_coarse_inter2,
				    real_d_coarse_restr2,
				    (void *) bary2};

