/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     Common/refine_common.c                                         */
/*                                                                          */
/*                                                                          */
/* description:  recursive refinement of a 2 and 3 dim. triangulations;     */
/*               file is included by 2d/refine.c and 3d/refine.c;           */
/*               file contains all routines not depending on the dimension; */
/*                                                                          */
/* global variables that have to be defined in 2d/refine.c and 3d/refine.c; */
/*  static MESH           *refine_mesh;                                     */
/*  static int            do_more_recursive_refine;                         */
/*  static int            new_coords;                                       */
/*  static int            call_refine_interpol;                             */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Univesitaet Bremen                                           */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*/
/* history:      first final version: 96/11/10                              */
/*                                    98/01/13  FE_SPACE structure          */
/*--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*/
/*  generate new coords on curved boundaries                                */
/*--------------------------------------------------------------------------*/

#if DIM == 3
static RC_LIST_EL *static_ref_list = nil;
#endif

#if DIM > 1
static void new_coords_fct(const EL_INFO *el_info)
{
  EL      *el = el_info->el;
  int     j;

#if DIM==2
  if (el->child[0] && el_info->boundary[2]
      && el_info->boundary[2]->param_bound && (el->new_coord == nil))
  {
    el->new_coord = get_real_d(el_info->mesh);
    for (j = 0; j < DIM_OF_WORLD; j++)
      el->new_coord[j] = (el_info->coord[0][j] + el_info->coord[1][j])*0.5;
    el_info->boundary[2]->param_bound(el->new_coord);
  }
#else
  int     i, n_neigh;
  DOF     *edge[2];
  EL_INFO *elinfo = (EL_INFO *) el_info;


  if (el->child[0] && el_info->boundary[N_FACES+0]
      && el_info->boundary[N_FACES+0]->param_bound && (el->new_coord == nil))
  {
    el->new_coord = get_real_d(el_info->mesh);
    for (j = 0; j < DIM_OF_WORLD; j++)
      el->new_coord[j] = (el_info->coord[0][j] + el_info->coord[1][j])*0.5;
    el_info->boundary[N_FACES+0]->param_bound(el->new_coord);

/*--------------------------------------------------------------------------*/
/*  now, information should be passed on to patch neighbours...             */
/*  get the refinement patch                                                */
/*--------------------------------------------------------------------------*/
    static_ref_list->el = el;
    static_ref_list->el_type = EL_TYPE(el, el_info);
    n_neigh = 1;
    for (i = 0; i < 2; i++)
      edge[i] = el_info->el->dof[i];
    if (get_refine_patch(&elinfo, edge, 0, static_ref_list, &n_neigh))
    {
/*--------------------------------------------------------------------------*/
/*  domain's boundary was reached while looping around the refinement edge  */
/*--------------------------------------------------------------------------*/
      get_refine_patch(&elinfo, edge, 1, static_ref_list, &n_neigh);
    }
    for (i = 1; i < n_neigh; i++)            /* start with 1, as list[0]=el */
    {
      TEST(static_ref_list[i].el->new_coord == nil)
	("non-nil new_coord in el %d ref_list[%d] el %d (n_neigh=%d)\n",
	 INDEX(el), i, INDEX(static_ref_list[i].el), n_neigh);

      static_ref_list[i].el->new_coord = el->new_coord;
    }
  }
#endif
  return;
}

/*--------------------------------------------------------------------------*/
/*  refine_fct: gets the refinement patch via get_refine_patch()            */
/*  and checks whether it compatible divisible (same refinement edge) or    */
/*  not; in the first case the patch is refined by refine_patch() in the    */
/*  second on the incompatible neighbour(s) is (are) refined first via a    */
/*  recursive call of recursive_refine_fct() in get_refine_patch().         */
/*--------------------------------------------------------------------------*/

static const EL_INFO  *refine_function(const EL_INFO *el_info)
{
  int          n_neigh, bound = false;
  DOF          *edge[2];
#if DIM == 2
  RC_LIST_EL   ref_list[2];
#else
  RC_LIST_EL   *ref_list;
#endif

  if (el_info->el->mark <= 0)  
    return(el_info);    /*   element may not be refined   */

#if DIM == 3
/*--------------------------------------------------------------------------*/
/*  get memory for a list of all elements at the refinement edge            */
/*--------------------------------------------------------------------------*/
  ref_list = get_rc_list(refine_mesh);
  ref_list->el_type = EL_TYPE(el_info->el,el_info);
#endif

  ref_list->el = el_info->el;
  n_neigh = 1;

/*--------------------------------------------------------------------------*/
/*  give the refinement edge the right orientation                          */
/*--------------------------------------------------------------------------*/

  if (el_info->el->dof[0][0] < el_info->el->dof[1][0])
  {
    edge[0] = el_info->el->dof[0];
    edge[1] = el_info->el->dof[1];
  }
  else
  {
    edge[1] = el_info->el->dof[0];
    edge[0] = el_info->el->dof[1];
  }

/*--------------------------------------------------------------------------*/
/*  get the refinement patch                                                */
/*--------------------------------------------------------------------------*/
  if (get_refine_patch((void *) &el_info, edge, 0, ref_list, &n_neigh))
  {
/*--------------------------------------------------------------------------*/
/*  domain's boundary was reached while looping around the refinement edge  */
/*--------------------------------------------------------------------------*/
    get_refine_patch((void *) &el_info, edge, 1, ref_list, &n_neigh);
    bound = true;
  }

#if DIM == 3
/*--------------------------------------------------------------------------*/
/*  fill neighbour information inside the patch in the refinement list      */
/*--------------------------------------------------------------------------*/
  get_neigh_on_patch(ref_list, n_neigh, bound);
#endif

/*--------------------------------------------------------------------------*/
/*  and now refine the patch                                                */
/*--------------------------------------------------------------------------*/
  bisect_patch(edge, ref_list, n_neigh, bound);

#if DIM == 3 
#if !NEIGH_IN_EL
  update_elinfo_stack(stack);
#endif
  free_rc_list(refine_mesh, ref_list);
#endif

  return(el_info);
}

#if NEIGH_IN_EL
static void refine_fct(const EL_INFO *el_info)
{
  FUNCNAME("refine_fct");

  refine_function((EL_INFO *) el_info);

  return;
}
#endif
#endif

/*--------------------------------------------------------------------------*/
/*  refine:     							    */
/*  traversal routine for recursiv refinement of a triangulation; bacic     */
/*  refinement module; used by all other refinement routines like           */
/*  global_refine()                                                         */
/*--------------------------------------------------------------------------*/

static int      count_refine_interpol(MESH *);
#if DIM == 1
U_CHAR refine(MESH *mesh)
{
  int           n_elements;

  n_elements = mesh->n_elements;
  call_refine_interpol = count_refine_interpol(mesh);

  do_more_refine = true;
  while (do_more_refine)
  {
    do_more_refine = false;
    mesh_traverse(mesh, -1, CALL_LEAF_EL, refine_fct);
  }
  n_elements = mesh->n_elements - n_elements;

  return(n_elements ? MESH_REFINED : 0);
}
#else
U_CHAR refine(MESH *mesh)
{
  int           n_elements;
#if !NEIGH_IN_EL
  const EL_INFO *el_info;
#endif

  refine_mesh = mesh;
  n_elements = mesh->n_elements;
  new_coords  = false;

  call_refine_interpol = count_refine_interpol(mesh);

#if NEIGH_IN_EL

  do_more_recursive_refine = true;
  while (do_more_recursive_refine)
  {
    do_more_recursive_refine = false;
    mesh_traverse(mesh, -1, CALL_LEAF_EL, refine_fct);
  }

#else /* NEIGH_IN_EL */
  stack = get_traverse_stack();
  do_more_recursive_refine = true;
  while (do_more_recursive_refine)
  {
    do_more_recursive_refine = false;
    el_info = traverse_first(stack, mesh, -1, 
			     CALL_LEAF_EL|FILL_NEIGH|FILL_BOUND);
    while (el_info)
    {
      if (el_info->el->mark > 0)
      {
	do_more_recursive_refine |= (el_info->el->mark > 1);
	el_info = refine_function(el_info);
      }
      el_info = traverse_next(stack, el_info);
    }
  }
#endif

  if (new_coords)
  {
    FLAGS fill_flag = CALL_EVERY_EL_PREORDER|FILL_BOUND|FILL_COORDS;
#if DIM == 2
    mesh_traverse(mesh, -1, fill_flag, new_coords_fct);
#else
    static_ref_list = get_rc_list(mesh);

#if NEIGH_IN_EL
    mesh_traverse(mesh, -1, fill_flag, new_coords_fct);
#else
    fill_flag |= FILL_NEIGH;
    el_info = traverse_first(stack, mesh, -1, fill_flag);
    while (el_info)
    {
      new_coords_fct(el_info);
      el_info = traverse_next(stack, el_info);
    }
#endif

    free_rc_list(mesh, static_ref_list);
    static_ref_list = nil;
#endif
  }

#if !NEIGH_IN_EL
  free_traverse_stack(stack);
#endif

  n_elements = mesh->n_elements - n_elements;

  return(n_elements ? MESH_REFINED : 0);
}
#endif

/*--------------------------------------------------------------------------*/
/*  refinement routines for global refinement:				    */
/*--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*/
/*  global_refine_fct:							    */
/*  sets the mark on all elements that have to be refined                   */
/*--------------------------------------------------------------------------*/

static int global_refine_mark;

static void global_refine_fct(const EL_INFO *el_info)
{
  el_info->el->mark = global_refine_mark;
}

/*--------------------------------------------------------------------------*/
/*  global_refine:							    */
/*  refines every element of mesh at least mark times			    */
/*--------------------------------------------------------------------------*/

U_CHAR global_refine(MESH *mesh, int mark)
{
  if (mark <= 0)  return(0);
  global_refine_mark = mark;
  mesh_traverse(mesh, -1, CALL_LEAF_EL, global_refine_fct);
  return(refine(mesh));
}

/*--------------------------------------------------------------------------*/
/* interpolation of dof vectors during refinement                           */
/* using DOF_VEC_LIST structure                                             */
/*--------------------------------------------------------------------------*/

typedef struct dof_vec_list  DOF_VEC_LIST;

struct dof_vec_list
{
  int             size;                       /* current size of total list */
  void            **list;          /* total list, distributed to ones below */

  int             n_dof_int_vec, n_dof_dof_vec, n_dof_uchar_vec,
                  n_dof_schar_vec, n_dof_real_vec, n_dof_real_d_vec,
                  n_dof_matrix;

  DOF_INT_VEC     **dof_int_vec;
  DOF_DOF_VEC     **dof_dof_vec;
  DOF_UCHAR_VEC   **dof_uchar_vec;
  DOF_SCHAR_VEC   **dof_schar_vec;
  DOF_REAL_VEC    **dof_real_vec;
  DOF_REAL_D_VEC  **dof_real_d_vec;
  DOF_MATRIX      **dof_matrix;
};

static DOF_VEC_LIST  dvlist[1] = {{0,nil}};

static int count_refine_interpol(MESH *mesh)
{
  FUNCNAME("count_refine_interpol");
  DOF_ADMIN      *admin;
  int            iadmin;
  DOF_REAL_VEC   *drv;
  DOF_REAL_D_VEC *drdv;
  DOF_INT_VEC    *div;
  DOF_DOF_VEC    *ddv;
  DOF_UCHAR_VEC  *duv;
  DOF_SCHAR_VEC  *dsv;
  DOF_MATRIX     *dm;
  int            nri = 0;
  int            n_dof_int_vec=0,  n_dof_dof_vec=0, n_dof_uchar_vec=0,
                 n_dof_schar_vec=0, n_dof_real_vec=0, n_dof_real_d_vec=0,
                 n_dof_matrix=0;

  for (iadmin = 0; iadmin < mesh->n_dof_admin; iadmin++)
  {
    admin = mesh->dof_admin[iadmin];

    for (div = admin->dof_int_vec; div; div = div->next)
    {
      if (div->refine_interpol) n_dof_int_vec++;
    }

    for (ddv = admin->dof_dof_vec; ddv; ddv = ddv->next)
    {
      if (ddv->refine_interpol) n_dof_dof_vec++;
    }

    for (ddv = admin->int_dof_vec; ddv; ddv = ddv->next)
    {
      if (ddv->refine_interpol) n_dof_dof_vec++;
    }

    for (duv = admin->dof_uchar_vec; duv; duv = duv->next)
    {
      if (duv->refine_interpol) n_dof_uchar_vec++;
    }

    for (dsv = admin->dof_schar_vec; dsv; dsv = dsv->next)
    {
      if (dsv->refine_interpol) n_dof_schar_vec++;
    }

    for (drv = admin->dof_real_vec; drv; drv = drv->next)
    {
      if (drv->refine_interpol) n_dof_real_vec++;
    }

    for (drdv = admin->dof_real_d_vec; drdv; drdv = drdv->next)
    {
      if (drdv->refine_interpol) n_dof_real_d_vec++;
    }

    for (dm = admin->dof_matrix; dm; dm = dm->next)
    {
      if (dm->refine_interpol) n_dof_matrix++;
    }
  }

  nri = n_dof_int_vec + n_dof_dof_vec + n_dof_uchar_vec + n_dof_schar_vec
    + n_dof_real_vec + n_dof_real_d_vec + n_dof_matrix;

  if (nri > 0)
  {
    if (dvlist->size < nri) {
      dvlist->list = MEM_REALLOC(dvlist->list, dvlist->size, nri+5, void *);
      dvlist->size = nri+5;
    }

    nri = 0;

    if (n_dof_int_vec)
      dvlist->dof_int_vec    = (DOF_INT_VEC **)(dvlist->list+nri);
    else
      dvlist->dof_int_vec    = nil;
    nri += n_dof_int_vec;

    if (n_dof_dof_vec)
      dvlist->dof_dof_vec    = (DOF_DOF_VEC **)(dvlist->list+nri);
    else
      dvlist->dof_dof_vec    = nil;
    nri += n_dof_dof_vec;

    if (n_dof_uchar_vec)
      dvlist->dof_uchar_vec  = (DOF_UCHAR_VEC **)(dvlist->list+nri);
    else
      dvlist->dof_uchar_vec  = nil;
    nri += n_dof_uchar_vec;

    if (n_dof_schar_vec)
      dvlist->dof_schar_vec  = (DOF_SCHAR_VEC **)(dvlist->list+nri);
    else
      dvlist->dof_schar_vec  = nil;
    nri += n_dof_schar_vec;

    if (n_dof_real_vec)
      dvlist->dof_real_vec   = (DOF_REAL_VEC **)(dvlist->list+nri);
    else
      dvlist->dof_real_vec   = nil;
    nri += n_dof_real_vec;

    if (n_dof_real_d_vec)
      dvlist->dof_real_d_vec = (DOF_REAL_D_VEC **)(dvlist->list+nri);
    else
      dvlist->dof_real_d_vec = nil;
    nri += n_dof_real_d_vec;

    if (n_dof_matrix)
      dvlist->dof_matrix     = (DOF_MATRIX **)(dvlist->list+nri);
    else
      dvlist->dof_matrix     = nil;
    nri += n_dof_matrix;
    TEST_EXIT(nri <= dvlist->size)("error in dvlist->size");

    dvlist->n_dof_int_vec    = 0;
    dvlist->n_dof_dof_vec    = 0;
    dvlist->n_dof_uchar_vec  = 0;
    dvlist->n_dof_schar_vec  = 0;
    dvlist->n_dof_real_vec   = 0;
    dvlist->n_dof_real_d_vec = 0;
    dvlist->n_dof_matrix     = 0;

    for (iadmin = 0; iadmin < mesh->n_dof_admin; iadmin++)
    {
      admin = mesh->dof_admin[iadmin];
      for (div = admin->dof_int_vec; div; div = div->next)
      {
	if (div->refine_interpol) 
	  dvlist->dof_int_vec[dvlist->n_dof_int_vec++] = div;
      }
      for (ddv = admin->dof_dof_vec; ddv; ddv = ddv->next) 
      {
	if (ddv->refine_interpol) 
	  dvlist->dof_dof_vec[dvlist->n_dof_dof_vec++] = ddv;
      }
      for (ddv = admin->int_dof_vec; ddv; ddv = ddv->next) 
      {
	if (ddv->refine_interpol) 
	  dvlist->dof_dof_vec[dvlist->n_dof_dof_vec++] = ddv;
      }
      for (duv = admin->dof_uchar_vec; duv; duv = duv->next) 
      {
	if (duv->refine_interpol)
	  dvlist->dof_uchar_vec[dvlist->n_dof_uchar_vec++] = duv;
      }
      for (dsv = admin->dof_schar_vec; dsv; dsv = dsv->next) 
      {
	if (dsv->refine_interpol) 
	  dvlist->dof_schar_vec[dvlist->n_dof_schar_vec++] = dsv;
      }
      for (drv = admin->dof_real_vec; drv; drv = drv->next) 
      {
	if (drv->refine_interpol)
	  dvlist->dof_real_vec[dvlist->n_dof_real_vec++] = drv;
      }
      for (drdv = admin->dof_real_d_vec; drdv; drdv = drdv->next) 
      {
	if (drdv->refine_interpol) 
	  dvlist->dof_real_d_vec[dvlist->n_dof_real_d_vec++] = drdv;
      }
      for (dm = admin->dof_matrix; dm; dm = dm->next) 
      {
	if (dm->refine_interpol) 
	  dvlist->dof_matrix[dvlist->n_dof_matrix++] = dm;
      }
    }

    TEST_EXIT(dvlist->n_dof_int_vec == n_dof_int_vec)
      ("error in n_dof_int_vec");
    TEST_EXIT(dvlist->n_dof_dof_vec == n_dof_dof_vec)
      ("error in n_dof_dof_vec");
    TEST_EXIT(dvlist->n_dof_uchar_vec == n_dof_uchar_vec)
      ("error in n_dof_uchar_vec");
    TEST_EXIT(dvlist->n_dof_schar_vec == n_dof_schar_vec)
      ("error in n_dof_schar_vec");
    TEST_EXIT(dvlist->n_dof_real_vec == n_dof_real_vec)
      ("error in n_dof_real_vec");
    TEST_EXIT(dvlist->n_dof_real_d_vec == n_dof_real_d_vec)
      ("error in n_dof_real_d_vec");
    TEST_EXIT(dvlist->n_dof_matrix == n_dof_matrix)
      ("error in n_dof_matrix");
  }
  else
  {
    dvlist->dof_int_vec    = nil;
    dvlist->dof_dof_vec    = nil;
    dvlist->dof_uchar_vec  = nil;
    dvlist->dof_schar_vec  = nil;
    dvlist->dof_real_vec   = nil;
    dvlist->dof_real_d_vec = nil;
    dvlist->dof_matrix     = nil;
  }

  return(nri);
}


static void refine_interpol(RC_LIST_EL *list, int n_el)
{
  DOF_REAL_VEC    *drv;
  DOF_REAL_D_VEC  *drdv;
  DOF_INT_VEC     *div;
  DOF_DOF_VEC     *ddv;
  DOF_UCHAR_VEC   *duv;
  DOF_SCHAR_VEC   *dsv;
  DOF_MATRIX      *dm;
  int             i;

  if (!dvlist) return;

  for (i = 0; i < dvlist->n_dof_int_vec; i++) 
  {
    div = dvlist->dof_int_vec[i];
    div->refine_interpol(div, list, n_el);
  }

  for (i = 0; i < dvlist->n_dof_dof_vec; i++) 
  {
    ddv = dvlist->dof_dof_vec[i];
    ddv->refine_interpol(ddv, list, n_el);
  }

  for (i = 0; i < dvlist->n_dof_uchar_vec; i++) 
  {
    duv = dvlist->dof_uchar_vec[i];
    duv->refine_interpol(duv, list, n_el);
  }

  for (i = 0; i < dvlist->n_dof_schar_vec; i++)
  {
    dsv = dvlist->dof_schar_vec[i];
    dsv->refine_interpol(dsv, list, n_el);
  }

  for (i = 0; i < dvlist->n_dof_real_vec; i++) 
  {
    drv = dvlist->dof_real_vec[i];
    drv->refine_interpol(drv, list, n_el);
  }

  for (i = 0; i < dvlist->n_dof_real_d_vec; i++) 
  {
    drdv = dvlist->dof_real_d_vec[i];
    drdv->refine_interpol(drdv, list, n_el);
  }

  for (i = 0; i < dvlist->n_dof_matrix; i++)
  {
    dm = dvlist->dof_matrix[i];
    dm->refine_interpol(dm, list, n_el);
  }

  return;
}
