/*--------------------------------------------------------------------------*/
/*  PLOT_ANSI: package for simple 2d X11 graphics                           */
/*                                                                          */
/*  (c) by A. Schmidt (1990-2004)                                           */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

/****************************************************************************
Header:

void PLOT_set_icolor(int icolor);
void PLOT_get_icolor(int *icolor);
void PLOT_set_backgroundcolor(float color[3]);
void PLOT_get_backgroundcolor(float color[3]);
void PLOT_set_linecolor(float color[3]);
void PLOT_get_linecolor(float color[3]);
void PLOT_set_patchcolor(float color[3]);
void PLOT_get_patchcolor(float color[3]);
void PLOT_set_textcolor(float color[3]);
void PLOT_get_textcolor(float color[3]);
void PLOT_flush(void);
void PLOT_clear(void);
void PLOT_set_transform2d(int mtype, float mat[3][3]);
void PLOT_get_transform2d(int mtype, float mat[3][3]);
void set_W2W_matrix2d(PLOT_INFO *info);
void PLOT_world2d(float xmin, float xmax, float ymin, float ymax);
void PLOT_move2d(float x, float y);
void PLOT_vmove2d(float vec[2]);
void PLOT_draw2d(float x, float y);
void PLOT_vdraw2d(float vec[2]);
void PLOT_line2d(float x1,float y1,float x2,float y2);
void PLOT_vline2d(float vec1[2], float vec2[2]);
void PLOT_triangle2d(float x1, float y1, float x2, float y2,
		     float x3, float y3);
void PLOT_vtriangle2d(float vec1[2], float vec2[2], float vec3[2]);
void PLOT_triangleRGB2d(float x1, float y1, float x2, float y2,
			float x3, float y3,
			float rgb1[3], float rgb2[3], float rgb3[3]);
void PLOT_vtriangleRGB2d(float vec1[2], float vec2[2], float vec3[2],
			 float rgb1[3], float rgb2[3], float rgb3[3]);
void PLOT_text2d(float x, float y, char *string);
void PLOT_vtext2d(float vec[2], char *string);

****************************************************************************/


/******************************************************************
  PLOT 2D package

  (with LOG)
******************************************************************/


#include <stdio.h>

#include <fcntl.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "PLOTansi.h"

extern PLOT_INFO *current_PLOT_info;
#define INFO current_PLOT_info


extern PLOT_LOG_ITEM *PLOT_LOG_get_item(void);

static PLOT_LOG_ITEM *LOG;


/******************************************************************
  2D World to Window coordinate transformation
******************************************************************/

static void World2d_to_Window(float worldcoord[2], float wincoord[3])
{
  wincoord[0] = worldcoord[0] * INFO->W2W_matrix[0][0]
               + worldcoord[1] * INFO->W2W_matrix[0][1]
		                + INFO->W2W_matrix[0][2];
  wincoord[1] = worldcoord[0] * INFO->W2W_matrix[1][0]
               + worldcoord[1] * INFO->W2W_matrix[1][1]
		                + INFO->W2W_matrix[1][2];
  wincoord[2] = -1.0;
}


/******************************************************************
  Color
******************************************************************/


#define PLOT_BackgroundColor 1
#define PLOT_LineColor 2
#define PLOT_PatchColor 3
#define PLOT_TextColor 4
#define PLOT_ScaleColor 5


static void test_set_color(float color[3], int colorType)
{
  if (INFO->current_color != colorType) {
    if ((LOG = PLOT_LOG_get_item())) {
      LOG->type = PLOT_LOG_color;
      LOG->U.f[0] = color[0];
      LOG->U.f[1] = color[1];
      LOG->U.f[2] = color[2];
    }
    INFO->current_color = colorType;
  }
}

static void set_color(const float color[3], int colorType)
{
  if ((LOG = PLOT_LOG_get_item())) {
    LOG->type = PLOT_LOG_color;
    LOG->U.f[0] = color[0];
    LOG->U.f[1] = color[1];
    LOG->U.f[2] = color[2];
  }
  INFO->current_color = colorType;
}

/************************ X11 color index *************/

void PLOT_set_icolor(int icolor)
{
  if ((LOG = PLOT_LOG_get_item())) {
    LOG->type = PLOT_LOG_icolor;
    LOG->U.i[0] = icolor;
  }
  INFO->icolor = icolor;
}

void PLOT_get_icolor(int *icolor)
{
  *icolor = INFO->icolor;
}

/************************ background color ************/

void PLOT_set_backgroundcolor(const float color[3])
{
  int i;
  for (i=0;i<3;i++) INFO->background_color[i] = color[i];
}

void PLOT_get_backgroundcolor(float color[3])
{
  int i;
  for (i=0;i<3;i++) color[i] = INFO->background_color[i];
}

/************************ line color ************/

void PLOT_set_linecolor(const float color[3])
{
  int i;
  for (i=0;i<3;i++) INFO->line_color[i] = color[i];
  set_color(color, PLOT_LineColor);
}

void PLOT_get_linecolor(float color[3])
{
  int i;
  for (i=0;i<3;i++) color[i] = INFO->line_color[i];
}

/************************ patch color ************/

void PLOT_set_patchcolor(const float color[3])
{
  int i;
  for (i=0;i<3;i++) INFO->patch_color[i] = color[i];
  set_color(color, PLOT_PatchColor);
}

void PLOT_get_patchcolor(float color[3])
{
  int i;
  for (i=0;i<3;i++) color[i] = INFO->patch_color[i];
}

/************************ text color ************/

void PLOT_set_textcolor(const float color[3])
{
  int i;
  for (i=0;i<3;i++) INFO->text_color[i] = color[i];
  set_color(color, PLOT_TextColor);
}

void PLOT_get_textcolor(float color[3])
{
  int i;
  for (i=0;i<3;i++) color[i] = INFO->text_color[i];
}




/******************************************************************
  Flush
******************************************************************/

void PLOT_flush(void)
{
  XEvent report;

  if (XCheckTypedWindowEvent(INFO->display, INFO->window, 
			Expose, &report))        {
    while (XCheckTypedWindowEvent(INFO->display, INFO->window, 
			     Expose, &report)) ;
    PLOT_expose(INFO->window);
  }
  else {
    /*printf("PLOT_flush() calls PLOT_LOG_flush(INFO)\n");*/
    PLOT_LOG_flush(INFO);
  }
}




/******************************************************************
  Clear
******************************************************************/

void PLOT_clear(void)
{
  free_log(INFO->log);
  INFO->log = NULL;
}




/******************************************************************
  Transformations
******************************************************************/

void PLOT_set_transform2d(int mtype, float mat[3][3])
{
  int i;
  float *smat = &mat[0][0];
  float *dmat;

  switch (mtype) {
  case PLOT_MODEL_MATRIX:
    dmat = &INFO->model_matrix[0][0];
    break;
  case PLOT_VIEW_MATRIX:
    dmat = &INFO->view_matrix[0][0];
    break;
  case PLOT_PROJECT_MATRIX:
    dmat = &INFO->project_matrix[0][0];
    break;
  default:
    return;
  }
  
  for (i=0;i<9;i++)  *dmat++ = *smat++;

  set_W2W_matrix2d(INFO);
}



void PLOT_get_transform2d(int mtype, float mat[3][3])
{
  int i;
  float *smat;
  float *dmat = &mat[0][0];

  switch (mtype) {
  case PLOT_MODEL_MATRIX:
    smat = &INFO->model_matrix[0][0];
    break;
  case PLOT_VIEW_MATRIX:
    smat = &INFO->view_matrix[0][0];
    break;
  case PLOT_PROJECT_MATRIX:
    smat = &INFO->project_matrix[0][0];
    break;
  default:
    return;
  }

  for (i=0;i<9;i++)  *dmat++ = *smat++;
}


void set_W2W_matrix2d(PLOT_INFO *info)
{
  int i,j,k;
  float mat1[3][3];
/*  int min_wh;*/

  /*  mat1 := view * model   */
  for (i=0;i<3;i++) {
    for (j=0;j<3;j++) {
      mat1[i][j] = 0.0;
      for (k=0;k<3;k++) {
	mat1[i][j] += info->view_matrix[i][k] * info->model_matrix[k][j];
      }
    }
  }
  

  /*  W2W = project * mat1  */
  for (i=0;i<3;i++) {
    for (j=0;j<3;j++) {
      info->W2W_matrix[i][j] = 0.0;
      for (k=0;k<3;k++) {
	info->W2W_matrix[i][j] += info->view_matrix[i][k] * mat1[k][j];
      }
    }
  }



  /*  now, look for window coordinates:  */
  /*  unit square has to fit into window, at lower left edge  */

/************  THIS WILL BE DONE WHEN LOG IS DISPLAYED ! ! !
  min_wh = info->windowWidth;
  if (min_wh > info->windowHeight)  min_wh = info->windowHeight;
  
  for (i=0;i<3;i++) {
    info->W2W_matrix[0][i] = info->W2W_matrix[0][i] * min_wh;
    info->W2W_matrix[1][i] = info->W2W_matrix[2][i] * info->windowHeight
                           - info->W2W_matrix[1][i] * min_wh;
  }
************/

}

/***************************************************************************/

void PLOT_world2d(float xmin, float xmax, float ymin, float ymax)
{
  float mat[3][3], xscal,yscal  /* ,xmid,ymid */;
  int i,j;

  xscal = xmax - xmin;
  if (xscal <= 0) {
    xscal = 1.0;
    fprintf(stderr,"PLOT_world2d: Invalid xmin/xmax values\n");
  }
  yscal = ymax - ymin;
  if (yscal <= 0) {
    yscal = 1.0;
    fprintf(stderr,"PLOT_world2d: Invalid ymin/ymax values\n");
  }

  mat[0][0] = 1.0 / xscal;
  mat[0][1] = 0.0;
  mat[0][2] = -xmin / xscal;
  mat[1][0] = 0.0;
  mat[1][1] = 1.0 / yscal;
  mat[1][2] = -ymin / yscal;
  mat[2][0] = 0.0;
  mat[2][1] = 0.0;
  mat[2][2] = 1.0;
  PLOT_set_transform2d(PLOT_MODEL_MATRIX, mat);

  for (i=0; i<3; i++)
    for (j=0; j<3; j++)
      mat[i][j] = (i==j ? 1.0 : 0.0);

  PLOT_set_transform2d(PLOT_VIEW_MATRIX, mat);
  PLOT_set_transform2d(PLOT_PROJECT_MATRIX, mat);
}


/******************************************************************
  Move
******************************************************************/

void PLOT_move2d(float x, float y)
{
  float vec[2];
  vec[0] = x; vec[1] = y;

  PLOT_vmove2d(vec);
}


void PLOT_vmove2d(float vec[2])
{
  if ((LOG = PLOT_LOG_get_item())) {
    LOG->type = PLOT_LOG_move;
    World2d_to_Window(vec,LOG->U.f);
  }
}



/******************************************************************
  Draw
******************************************************************/

void PLOT_draw2d(float x, float y)
{
  float vec[2];
  vec[0] = x; vec[1] = y;

  PLOT_vdraw2d(vec);
}


void PLOT_vdraw2d(float vec[2])
{
  test_set_color(INFO->line_color, PLOT_LineColor);
  if ((LOG = PLOT_LOG_get_item())) {
    LOG->type = PLOT_LOG_draw;
    World2d_to_Window(vec,LOG->U.f);
  }
}


/******************************************************************
  Line
******************************************************************/

void PLOT_line2d(float x1,float y1,float x2,float y2)
{
  float vec1[2],vec2[2];
  vec1[0] = x1; vec1[1] = y1;
  vec2[0] = x2; vec2[1] = y2;

  PLOT_vline2d(vec1,vec2);
}


void PLOT_vline2d(float vec1[2], float vec2[2])
{
  test_set_color(INFO->line_color, PLOT_LineColor);

  if ((LOG = PLOT_LOG_get_item())) {
    LOG->type = PLOT_LOG_move;
    World2d_to_Window(vec1,LOG->U.f);
  }
  if ((LOG = PLOT_LOG_get_item())) {
    LOG->type = PLOT_LOG_draw;
    World2d_to_Window(vec2,LOG->U.f);
  }
}


/******************************************************************
  triangle
******************************************************************/

void PLOT_triangle2d(float x1, float y1, float x2, float y2,
		     float x3, float y3)
{
  float vec1[2],vec2[2],vec3[2];
  vec1[0] = x1; vec1[1] = y1;
  vec2[0] = x2; vec2[1] = y2;
  vec3[0] = x3; vec3[1] = y3;

  PLOT_vtriangle2d(vec1,vec2,vec3);
}


void PLOT_vtriangle2d(float vec1[2], float vec2[2], float vec3[2])
{
  test_set_color(INFO->patch_color, PLOT_PatchColor);

  if ((LOG = PLOT_LOG_get_item())) {
    LOG->type = PLOT_LOG_triang;
    World2d_to_Window(vec1,LOG->U.f);
  }
  if ((LOG = PLOT_LOG_get_item())) {
    LOG->type = PLOT_LOG_triang;
    World2d_to_Window(vec2,LOG->U.f);
  }
  if ((LOG = PLOT_LOG_get_item())) {
    LOG->type = PLOT_LOG_triang;
    World2d_to_Window(vec3,LOG->U.f);
  }
}



void PLOT_triangleRGB2d(float x1, float y1, float x2, float y2,
			float x3, float y3,
			float rgb1[3], float rgb2[3], float rgb3[3])
{
  float vec1[2],vec2[2],vec3[2];
  vec1[0] = x1; vec1[1] = y1;
  vec2[0] = x2; vec2[1] = y2;
  vec3[0] = x3; vec3[1] = y3;

  PLOT_vtriangleRGB2d(vec1,vec2,vec3, rgb1,rgb2,rgb3);
}


void PLOT_vtriangleRGB2d(float vec1[2], float vec2[2], float vec3[2],
			 float rgb1[3], float rgb2[3], float rgb3[3])
{
  int i;
  static float White[3] = {1.0,1.0,1.0};

  test_set_color(White, PLOT_ScaleColor);

  if ((LOG = PLOT_LOG_get_item())) {
    LOG->type = PLOT_LOG_triangRGB;
    World2d_to_Window(vec1,LOG->U.f);
  }
  if ((LOG = PLOT_LOG_get_item())) {
    LOG->type = PLOT_LOG_triangRGB;
    for (i=0;i<3;i++) LOG->U.f[i] = rgb1[i];
  }
  if ((LOG = PLOT_LOG_get_item())) {
    LOG->type = PLOT_LOG_triangRGB;
    World2d_to_Window(vec2,LOG->U.f);
  }
  if ((LOG = PLOT_LOG_get_item())) {
    LOG->type = PLOT_LOG_triangRGB;
    for (i=0;i<3;i++) LOG->U.f[i] = rgb2[i];
  }
  if ((LOG = PLOT_LOG_get_item())) {
    LOG->type = PLOT_LOG_triangRGB;
    World2d_to_Window(vec3,LOG->U.f);
  }
  if ((LOG = PLOT_LOG_get_item())) {
    LOG->type = PLOT_LOG_triangRGB;
    for (i=0;i<3;i++) LOG->U.f[i] = rgb3[i];
  }
}



/******************************************************************
  text
******************************************************************/


void PLOT_text2d(float x, float y, char *string)
{
  float vec[2];
  vec[0] = x;
  vec[1] = y;
  PLOT_vtext2d(vec,string);
}



void PLOT_vtext2d(float vec[2], char *string)
{
  int i,len;
  
  test_set_color(INFO->text_color, PLOT_TextColor);

  if ((LOG = PLOT_LOG_get_item())) {
    LOG->type = PLOT_LOG_text;
    World2d_to_Window(vec,LOG->U.f);
  }
  
  len = strlen(string);
  while (len>0) {
    if ((LOG = PLOT_LOG_get_item())) {
      LOG->type = PLOT_LOG_string;
      for (i=0;i<PLOT_LOG_charlen; i++) {
	LOG->U.c[i] = *string++;
	if(--len <= 0) break;
      }
      /* add a final NUL, if there is place for it */
      if ((len <= 0) && (i < PLOT_LOG_charlen-1)) LOG->U.c[i+1] = 0;
    }
    else break;
  }
}
